#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a GLib plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; The Toolkit Thread
;;; package: (glib thread)
;;; parent: (runtime thread)

(define glib-thread #f)

;;; With the glib-thread always running, the runtime system should no
;;; longer use wait-for-io and thus never signal
;;; condition-type:no-thread!

;;; GC daemons cannot be allowed to run during a callback.  After-gc
;;; interrupts are currently serviced with interrupt-mask/timer-ok!,
;;; which might allow a switch to a different thread, which might
;;; return from a different callback.

;;; The Glib system's "GC cleanups" are run by glib-thread sometime
;;; after a flip.  The secondary gc daemons are also run by glib-thread
;;; after some number of flips.

;; Number of GCs between applications of trigger-secondary-gc-daemons!
(define secondary-gc-rate 100)

(define (create-glib-thread)
  (if glib-thread (error "A Glib thread already exists."))
  (set! glib-thread
	(create-thread
	 #f (lambda ()
	      (let ((self (current-thread))
		    (done-tick 0)
		    (next-secondary-tick secondary-gc-rate))
		(let glib-thread-loop ()
		  (with-glib-lock
		   (lambda ()
		     (let ((gc-tick (car (gc-timestamp))))
		       (if (fix:< done-tick gc-tick)
			   (begin
			     (%trace ";run-glib cleaning up\n")
			     (run-glib-cleanups)
			     (%trace ";run-glib clean up done\n")
			     (set! done-tick gc-tick)))
		       (if (fix:< next-secondary-tick gc-tick)
			   (begin
			     (%trace ";run-glib secondary-gc daemons\n")
			     (run-glib-daemons)
			     (%trace ";run-glib secondary-gc daemons done\n")
			     (set! next-secondary-tick
				   (fix:+ gc-tick secondary-gc-rate)))))
		     (with-thread-timer-stopped
		      (lambda ()
			(let ((time (or (and (thread/next self) 0)
					(and timer-records
					     (timer-record/time timer-records))
					-1)))
			  (%trace ";run-glib until "time"\n")
			  ;;(account-for-times self (get-system-times))
			  (run-glib (select-registry-handle io-registry) time)
			  ;;(record-start-times! self)
			  (%trace ";run-glib done at "(real-time-clock)"\n")
			  )))))
		  (%trace ";run-glib yields\n")
		  (yield-current-thread)
		  (%trace ";run-glib loops\n")
		  (glib-thread-loop))))
	 'glib))
  (detach-thread glib-thread))

(define (exit-glib-thread)
  (let ((thread glib-thread))
    (set! glib-thread #f)
    (if (not thread) (error "A Glib thread was not running."))
    (signal-thread-event
     thread (lambda ()
	      (exit-current-thread #t)))))

(define (stop-glib-thread)
  (signal-thread-event glib-thread (lambda () (stop-current-thread))))

(define (restart-glib-thread)
  (restart-thread glib-thread #t #f))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))