#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a GLib plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Core utilities.
;;; package: (glib internal)

(define-syntax define-integrable-operator
  #;(er-macro-transformer
   (lambda (form rename compare)
     (declare (ignore compare))
     (cond ((syntax-match? '((identifier . mit-bvl) + form) (cdr form))
	    (let ((r-begin (rename 'begin))
		  (r-declare (rename 'declare))
		  (r-define (rename 'define)))
	      `(,r-begin
		(,r-declare (integrate-operator ,(caadr form)))
		(,r-define ,@(cdr form)))))
	   (else
	    (ill-formed-syntax form)))))

  (rsc-macro-transformer
   (lambda (form environment)
     (declare (ignore environment))
     (if (syntax-match? '((identifier . mit-bvl) + form) (cdr form))
	 `(begin
	   (declare (integrate-operator ,(caadr form)))
	   (define ,@(cdr form)))
	 (ill-formed-syntax form)))))

(define-syntax error-if-null
  (syntax-rules ()
    ((_ alien message ...)
     (if (alien-null? alien)
	 (error message ...)))))

(define-integrable-operator (fix:max n m) (if (fix:> n m) n m))

(define-integrable-operator (fix:min n m) (if (fix:< n m) n m))

(define-integrable-operator (fix:negate i) (fix:- 0 i))

(define-integrable-operator (fix:abs n)
  (if (fix:negative? n) (fix:negate n) n))

(define-integrable (bit-mask-indices num)
  ;; The indices of the bits set in num.
  (let ((str (unsigned-integer->bit-string 32 num)))
    (let loop ((start 0))
      (let ((next (bit-substring-find-next-set-bit str start 32)))
	(if next
	    (cons next (loop (fix:1+ next)))
	    '())))))

(define-integrable (bit? int mask)
  (not (int:zero? (bitwise-and int mask))))

(declare (integrate-operator bit-ior))
(define (bit-ior . ints)
  (reduce bitwise-ior 0 ints))

;;; GLib Mutex

(define glib-mutex)

(define (reset-glib-mutex!)
  (set! glib-mutex (make-thread-mutex)))

(define (with-glib-lock thunk)
  (with-thread-mutex-lock glib-mutex thunk))

(define (without-glib-lock thunk)
  (without-thread-mutex-lock glib-mutex thunk))

(define (assert-glib-locked operator)
  ;; Useful at least when debugging single threaded worlds.
  (if (not (eq? (current-thread) (thread-mutex-owner glib-mutex)))
      (outf-error ";glib not locked: "operator"\n")))

(define (assert-without-interruption operator)
  (if (not (get-thread-event-block))
      (outf-error ";not without-interruption: "operator"\n")))

;;; GLib Cleanups

(define glib-cleanups)

(define (run-glib-cleanups)
  (assert-glib-locked 'run-glib-cleanups)
  (let loop ((alist glib-cleanups)
	     (prev #f))
    (if (pair? alist)
	(if (weak-pair/car? (car alist))
	    (loop (cdr alist) alist)
	    (let ((thunk (weak-cdr (car alist)))
		  (next (cdr alist)))
	      (with-glib-error-handler
	       "this GLib cleanup"
	       (lambda () (without-interruption thunk)))
	      (if prev
		  (set-cdr! prev next)
		  (set! glib-cleanups next))
	      (loop next prev))))))

(define (run-glib-daemons)
  (assert-glib-locked 'run-glib-daemons)
  (with-glib-error-handler "the secondary GC daemons"
			   trigger-secondary-gc-daemons!))

(define (with-glib-error-handler what thunk)
  (call-with-current-continuation
   (lambda (continue)
     ;;(with-restart name reporter effector interactor thunk)
     (with-restart 'abort
	 (string-append "Punt "what"; return to the GLib main loop.")
	 (named-lambda (glib-abort-effector . args)
	   (declare (ignore args))
	   (continue unspecific))
	 (named-lambda (glib-abort-interactor . args)
	   (apply values args))
       thunk))))

(define (reset-glib-cleanups!)
  (set! glib-cleanups '()))

(define (add-glib-cleanup object cleanup-thunk)
  (assert-glib-locked 'add-glib-cleanup)
  (let ((weak-pair (weak-cons object cleanup-thunk)))
    (set! glib-cleanups (cons weak-pair glib-cleanups))
    weak-pair))

(define (execute-glib-cleanup object)
  (assert-without-interruption 'execute-glib-cleanup)
  (assert-glib-locked 'execute-glib-cleanup)
  (let ((entry (weak-assq object glib-cleanups)))
    (if entry
	(begin
	  ((weak-cdr entry))
	  (set! glib-cleanups (delq! entry glib-cleanups)))))
  unspecific)

(define (weak-assq obj alist)
  (let loop ((alist alist))
    (if (null? alist) #f
	(let* ((entry (car alist))
	       (key (weak-car entry)))
	  (if (eq? obj key) entry
	      (loop (cdr alist)))))))

(define (reset-glib-package!)
  (reset-glib-mutex!)
  (reset-glib-cleanups!))

(define (initialize-package!)
  (reset-glib-package!)
  (add-event-receiver! event:after-restore reset-glib-package!))

(initialize-package!)