/* -*-C-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a GLib plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

*/

/* SchemeSource -- the custom GSource that runs Scheme in an idle task. */

#include <mit-scheme.h>
#include <glib.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <malloc.h>

/* Presumed externs/const of the Glib-ready machine. */
extern double OS_real_time_clock (void);
extern int OS_process_any_status_change (void);
extern int OS_select_registry_length (unsigned long registry);
#define SELECT_MODE_READ 1
#define SELECT_MODE_WRITE 2
extern void OS_select_registry_entry (unsigned long registry,
				      int i, int *fd, unsigned int *mode);
extern void OS_syserr_names (unsigned long *, const char ***);
extern void foreach_async_signal (void(*func)(int signo));
extern void abort_to_c (void);
extern void re_enter_scheme (void);
extern int interrupts_p (void);

static void init_signal_handling (void);

struct _SchemeSource
{
  GSource source;

  /* The list of GPollFDs that have been added to the main_context. */
  GSList * gpollfds;

  /* When to give up waiting. */
  double time_limit;

  /* TRUE when Scheme has a runnable thread.  Set to FALSE at the
     start of run_glib.  Set to TRUE by a callback that has made a
     Scheme thread runnable.  */
  gboolean runnable;
};
typedef struct _SchemeSource SchemeSource;

static gboolean scheme_source_prepare (GSource * source, gint * timeout);
static gboolean scheme_source_check (GSource * source);
static int pending_io (SchemeSource * source);
static gboolean scheme_source_dispatch (GSource * source, GSourceFunc callback, gpointer user_data);
static void install_scheme_source (void);
static void destroy_scheme_source (void);
static void clear_registry (SchemeSource * source);
static void set_registry (SchemeSource * source, GSList * new, double time);

static SchemeSource * scheme_source = NULL;
static gboolean tracing_glib_select = 0;
static void trace (const char *format, ...);
static GSList * glib_registry (unsigned long registry);

int slice_counter = 0;
static gchar * gpollfds_string (GSList * gpollfds);

void
trace (const char * format, ...)
{
  va_list args;
  va_start (args, format);
  if (tracing_glib_select)
    {
      vfprintf (stderr, format, args);
      fflush (stderr);
    }
  va_end (args);
}

static gboolean
scheme_source_prepare (GSource * source, gint * timeout)
{
  /* Return TRUE when ready to dispatch (without a poll).

     Return FALSE and set `timeout' to do a poll/check before
     dispatching. */

  SchemeSource * src = (SchemeSource *)source;

  if (src->runnable
      || interrupts_p ()
      || OS_process_any_status_change ())
    {
      trace (";scheme_source_prepare: ready (%s)\n",
	     src->runnable ? "thread"
	     : interrupts_p () ? "interrupt"
	     : "subprocess");
      *timeout = 0;
      return (TRUE);
    }
  if (src->time_limit == -1.0)
    {
      trace (";scheme_source_prepare: waiting\n");
      *timeout = -1;
      return (FALSE);
    }
  if (src->time_limit == 0.0)
    {
      trace (";scheme_source_prepare: polling\n");
      *timeout = 0;
      return (FALSE);
    }
  {
    double dtime = OS_real_time_clock ();
    gint timeo = ceil (src->time_limit - dtime);

    if (timeo <= 0)
      {
	trace (";scheme_source_prepare: ready (timeout)\n");
	*timeout = 0;
	return (TRUE);
      }

    trace (";scheme_source_prepare: polling for %dmsec\n", timeo);
    *timeout = timeo;
    return (FALSE);
  }
}

static gboolean
scheme_source_check (GSource * source)
{
  /* Return TRUE when ready to dispatch (after the poll). */

  SchemeSource * src = (SchemeSource *)source;

  if (src->time_limit == 0.0
      || src->runnable
      || interrupts_p ()
      || OS_process_any_status_change ()
      || pending_io (src))
    {
      trace (";scheme_source_check: ready (%s)\n",
	     src->runnable ? "thread"
	     : interrupts_p () ? "interrupt"
	     : OS_process_any_status_change () ? "subprocess"
	     : src->time_limit == 0.0 ? "" : "io");
      return (TRUE);
    }
  if (src->time_limit == -1.0)
    {
      trace (";scheme_source_check: waiting forever\n");
      return (FALSE);
    }
  {
    double dtime = OS_real_time_clock ();
    gint timeo = ceil (src->time_limit - dtime);

    if (timeo <= 0)
      {
	trace (";scheme_source_check: ready (timeout)\n");
	return (TRUE);
      }

    trace (";scheme_source_check: waiting %dmsec\n", timeo);
    return (FALSE);
  }
}

static int
pending_io (SchemeSource * src)
{
  GSList * scan;

  if (tracing_glib_select)
    {
      scan = src->gpollfds;
      while (scan != NULL)
	{
	  GPollFD * gfd = scan->data;
	  if (gfd->revents != 0)
	    {
	      fprintf (stderr, ";scheme_source_check: io ready on %d\n",
		       gfd->fd);
	    }
	  scan = scan->next;
	}
    }

  scan = src->gpollfds;
  while (scan != NULL)
    {
      GPollFD * gfd = scan->data;
      if (gfd->revents != 0)
	return (TRUE);
      scan = scan->next;
    }
  return (FALSE);
}

static gboolean
scheme_source_dispatch (GSource * source,
			GSourceFunc callback, gpointer user_data)
{
  /* Executes our "idle" task.  Ignore the callback and user_data
     arguments.  Must return TRUE to stay on the list of event
     sources. */

  if (g_source_is_destroyed (source))
    return (FALSE);

  slice_counter += 1;
  trace (";scheme_source_dispatch: running time slice %d\n", slice_counter);

  re_enter_scheme ();

  trace (";scheme_source_dispatch: finished time slice %d\n", slice_counter);
  return (TRUE);		/* Not a once-only. */
}

GSourceFuncs scheme_source_funcs =
{
  scheme_source_prepare,
  scheme_source_check,
  scheme_source_dispatch,
  NULL,
  NULL,
  NULL
};

static void
install_scheme_source (void)
{
  scheme_source = (SchemeSource *)
    g_source_new (&scheme_source_funcs, sizeof (SchemeSource));
  scheme_source->gpollfds = NULL;
  scheme_source->time_limit = 0.0;
  scheme_source->runnable = FALSE;
  g_source_set_priority ((GSource *) scheme_source, G_PRIORITY_LOW);
  g_source_attach ((GSource *) scheme_source, NULL);
}

static void
destroy_scheme_source (void)
{
  clear_registry (scheme_source);
  g_source_destroy ((GSource *) scheme_source);
  scheme_source = NULL;
}

static void
clear_registry (SchemeSource * source)
{
  GSList * gpollfds = source->gpollfds;
  if (gpollfds != NULL)
    {
      GMainContext * context = g_source_get_context ((GSource *)source);
      GSList * scan = gpollfds;
      while (scan != NULL)
	{
	  GPollFD * gfd = scan->data;
	  g_main_context_remove_poll (context, gfd);
	  g_free (gfd);
	  scan->data = NULL;
	  scan = scan->next;
	}
      g_slist_free (gpollfds);
    }
  source->gpollfds = NULL;
}

static void
set_registry (SchemeSource * source, GSList * new, double time)
{
  /* Set the source's current gpollfds to match NEW.  Warns if the
     registry is already set. */

  if (source->gpollfds != NULL)
    clear_registry (source);

  source->time_limit = time;
  source->runnable = FALSE;
  source->gpollfds = new;
  {
    GMainContext * context = g_source_get_context ((GSource *)source);
    while (new != NULL)
      {
	GPollFD * gfd = new->data;
	/* G_PRIORITY_LOW ensures that window resizes can happen even
	   when Scheme is spinning, thus allowing the time-slice
	   window to grow with its count. */
	g_main_context_add_poll (context, gfd, G_PRIORITY_LOW);
	new = new->next;
      }
  }
}


/* Invoking g_main_loop_run. */

extern SCM Scm_continue_start_glib (void);
extern SCM Scm_continue_stop_glib (void);
extern int cstack_depth;
typedef void (*SliceHook)(void);
SliceHook slice_hook = NULL;
static GMainLoop *loop;

gboolean
start_glib (void)
{
  /* Runs g_main_loop_run with scheme_source attached.  Returns TRUE when
     successful.  Returns FALSE when main loop is already running. */
  CalloutTrampIn tramp = &Scm_continue_start_glib;
  gboolean retval = TRUE;

  if (scheme_source != NULL)
    return (FALSE);

  slice_hook = NULL;
  init_signal_handling ();
  install_scheme_source ();

  callout_unseal (tramp);
  CSTACK_PUSH (gboolean, retval);
  CSTACK_PUSH (int, cstack_depth);
  CSTACK_PUSH (CalloutTrampIn, tramp);

  loop = g_main_loop_new (NULL, TRUE);
  g_main_loop_run (loop);
  g_main_loop_unref (loop);
  destroy_scheme_source ();
  return (FALSE);
}

void
stop_glib (void)
{
  /* Returns TRUE when successful. */

  if (scheme_source == NULL)
    return;
  g_main_loop_quit (loop);
  /* NOTREACHED */
}

void
run_glib (unsigned long registry, double time)
{
  /* Return to the toolkit with the scheme_source set up to dispatch
     to Scheme again when IO is ready, or a certain TIME has passed.
     If TIME has already passed, the IO registry is ignored and
     Scheme is ready to run again immediately.  If IO is empty, the
     simulated poll should not re-enter Scheme until TIME. */

  set_registry (scheme_source,
		glib_registry (registry),
		time);
  if (tracing_glib_select)
    {
      GSList * gpollfds = scheme_source->gpollfds;
      gchar * fdstr = gpollfds_string (gpollfds);
      fprintf (stderr, ";run_glib%s%s until %.1f\n",
	       gpollfds == NULL ? "" : " waiting on", fdstr, time);
      fflush (stderr);
      if (fdstr[0] != '\0')
	g_free (fdstr);
    }

  if (slice_hook != NULL) (*slice_hook)();

  abort_to_c ();
  /*NOTREACHED*/
}

void
yield_glib (void)
{
  scheme_source->runnable = TRUE;
  trace (";yield_glib: runnable at %.1f\n", OS_real_time_clock ());
}

/* Glib Select Registries -- GSLists of GPollFDs. */

/* SELECT_MODE_ -> GIOCondition */
#define DECODE_MODE(mode)						\
 (((((mode) & SELECT_MODE_READ) != 0) ? G_IO_IN : 0)			\
  | ((((mode) & SELECT_MODE_WRITE) != 0) ? G_IO_OUT : 0))

/* GIOCondition -> SELECT_MODE_ */
#define ENCODE_MODE(revents)						\
 (((((revents) & G_IO_IN) != 0) ? SELECT_MODE_READ : 0)			\
  | ((((revents) & G_IO_OUT) != 0) ? SELECT_MODE_WRITE : 0)		\
  | ((((revents) & G_IO_ERR) != 0) ? SELECT_MODE_ERROR : 0)		\
  | ((((revents) & G_IO_HUP) != 0) ? SELECT_MODE_HUP : 0))

static GSList *
glib_registry (unsigned long registry)
{
  /* Construct Glib's version of a select_registry_t. */

  int len = OS_select_registry_length (registry);
  int i = 0;
  GSList * list = NULL;

  while (i < len)
    {
      int fd;
      unsigned int mode;
      GPollFD * item = g_malloc (sizeof (GPollFD));
      OS_select_registry_entry (registry, i, (&fd), (&mode));
      item->fd = fd;
      item->events = DECODE_MODE (mode) | G_IO_ERR | G_IO_HUP;
      item->revents = 0;
      list = g_slist_prepend (list, item);
      i += 1;
    }
  return (list);
}

static gchar *
gpollfds_string (GSList * gpollfds)
{
  /* Construct a string describing the fds and r/w flags in GPOLLFDS,
     e.g. " 0(r)" */

  gchar * string = NULL;
  GSList * scan = gpollfds;
  while (scan != NULL)
    {
      GPollFD * gfd = scan->data;
      int mode = (gfd->events) & (~(G_IO_HUP|G_IO_ERR));
      gchar * next = g_strdup_printf ("%s %d(%s)",
				      (string == NULL ? "" : string),
				      gfd->fd,
				      (mode == (G_IO_IN|G_IO_OUT) ? "rw"
				       : mode == G_IO_IN ? "r"
				       : mode == G_IO_OUT ? "w" : "?"));
      if (string != NULL)
	g_free (string);
      string = next;
      scan = scan->next;
    }
  return (string);
}

gchar *
current_gpollfds_string (void)
{
  return (gpollfds_string (scheme_source->gpollfds));
}

gboolean
glib_select_trace_p (void)
{
  return (tracing_glib_select);
}

void
glib_select_trace (gboolean trace_p)
{
  tracing_glib_select = trace_p;
}

/* signal_forwarder

   This signal handler can run in any thread because it forwards the
   signal to the scheme_thread.  When the handler (subsequently) finds
   itself running in the scheme_thread, it invokes the original
   handler. */

#include <signal.h>
#include <pthread.h>
static const char * errno_name (int err);
static void complain (const char *format, ...);

static pthread_t scheme_thread;
static struct handler_record * old_handlers = NULL;
struct handler_record
{
  int signo;
  void (*handler)(int, siginfo_t *, void *);
  struct handler_record *next;
};

void
signal_forwarder (int signo, siginfo_t *siginfo, void *ptr)
{
  pthread_t self;

  self = pthread_self ();
  if (self == scheme_thread)
    {
      struct handler_record * scan;

      trace (";signal %d: in scheme_thread\n", signo);
      scan = old_handlers;
      while (scan != NULL)
	{
	  if (scan->signo == signo)
	    {
	      trace (";signal %d: invoking original handler\n", signo);
	      (scan->handler)(signo, siginfo, ptr);
	      trace (";signal %d: original handler returned\n", signo);
	      return;
	    }
	  scan = scan->next;
	}
      complain (";signal %d: no handler\n", signo);
    }
  else
    {
      int err;

      trace (";signal %d: outside scheme_thread\n", signo);
      err = pthread_kill (scheme_thread, signo);
      if (err != 0)
	{
	  complain (";signal %d: pthread_kill failed: %s\n",
		    signo, errno_name (err));
	  sleep (1);
	}
    }
}

static void
init_signal_forwarder (int signo)
{
  int err;
  struct handler_record *hrec;
  struct sigaction act;

  err = sigaction (signo, 0, (&act));
  if (err != 0)
    {
      complain ("init_signal_forwarder: sigaction access failed\n");
      return;
    }

  if (((act.sa_flags & SA_SIGINFO) == 0)
      && ((act.sa_handler == SIG_DFL)
	  || (act.sa_handler == SIG_IGN)))
    return;

  if ((act.sa_flags & SA_SIGINFO) == 0)
    {
      complain ("init_signal_forwarder: no SA_SIGINFO\n");
      return;
    }

  hrec = malloc (sizeof (struct handler_record));
  if (hrec == NULL)
    {
      complain ("init_signal_forwarder: malloc failed\n");
      return;
    }
  hrec->signo = signo;
  hrec->handler = act.sa_sigaction;
  hrec->next = old_handlers;
  act.sa_sigaction = &signal_forwarder;
  err = sigaction (signo, &act, 0);
  if (err != 0)
    complain ("init_signal_forwarder: sigaction modify failed\n");
  old_handlers = hrec;
}

static void
init_signal_handling (void)
{
  scheme_thread = pthread_self ();
  foreach_async_signal (&init_signal_forwarder);
}

static const char *
errno_name (int err)
{
  unsigned long length;
  const char ** names;
  OS_syserr_names (&length, &names);
  if (err < length)
    return names[err];
  else
    return "unknown errno";
}

static void
complain (const char *format, ...)
{
  va_list args;
  va_start (args, format);
  vfprintf (stderr, format, args);
  fflush (stderr);
  va_end (args);
}
